/*
 * Copyright (c) 2013-2014 Intel Corporation.  All rights reserved.
 *
 * This software is available to you under the BSD license
 * below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AWV
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <time.h>
#include <netdb.h>
#include <unistd.h>

#include <rdma/fabric.h>
#include <rdma/fi_errno.h>
#include <rdma/fi_endpoint.h>
#include <rdma/fi_cm.h>
#include <shared.h>

// MULTI_BUF_SIZE_FACTOR defines how large the multi recv buffer will be.
// The minimum value of the factor is 2 which will set the multi recv buffer
// size to be twice the size of the send buffer. In order to use FI_MULTI_RECV
// feature efficiently, we need to have a large recv buffer so that we don't
// to repost the buffer often to get the remaining data when the buffer is full
#define MULTI_BUF_SIZE_FACTOR 4
#define DEFAULT_MULTI_BUF_SIZE 1024*1024
#define SYNC_DATA_SIZE 16


static int max_credits = 128;
static char test_name[10] = "custom";
static struct timespec start, end;
static void *send_buf, *multi_recv_buf;
static size_t max_send_buf_size, multi_buf_size;

static struct fid_mr *mr_multi_recv;;
static void *local_addr, *remote_addr;
static size_t addrlen = 0;
static fi_addr_t remote_fi_addr;
struct fi_ctx_multi_recv *ctx_send;
struct fi_context fi_ctx_av;
struct fi_ctx_multi_recv *ctx_multi_recv;

struct fi_ctx_multi_recv {
	struct fi_context context;
	int index;
};

enum data_type {
	DATA = 1,
	CONTROL
};

int wait_for_send_completion(int num_completions)
{
	int ret;
	struct fi_cq_data_entry comp = {0};

	while (num_completions > 0) {
	 	memset(&comp, 0, sizeof(comp));
		ret = fi_cq_read(txcq, &comp, 1);
		if (ret > 0) {
			num_completions--;
		} else if (ret < 0 && ret != -FI_EAGAIN) {
			FT_PRINTERR("fi_cq_read", ret);
			return ret;
		}
	}
	return 0;
}

int wait_for_recv_completion(void **recv_data, enum data_type type,
		int num_completions)
{
	int ret;
	struct fi_cq_data_entry comp = {0};
	struct fi_ctx_multi_recv *header;
	int buf_index;

	while (num_completions > 0) {
	 	memset(&comp, 0, sizeof(comp));
		ret = fi_cq_read(rxcq, &comp, 1);
		if (ret > 0) {
 			header = (struct fi_ctx_multi_recv *)comp.op_context;
			buf_index = header->index;
			// buffer is used up, post it again
			if(comp.flags & FI_MULTI_RECV) {
				ret = fi_recv(ep, multi_recv_buf +
						(multi_buf_size/2) * buf_index,
						multi_buf_size/2,
						fi_mr_desc(mr_multi_recv),
						remote_fi_addr,
						&ctx_multi_recv[buf_index].context);
				if (ret) {
					FT_PRINTERR("fi_recv", ret);
					return ret;
				}
			} else {
				num_completions--;
				// check the completion event is for parameter
				// exchanging otherwise, do nothing for sync
				// data and multi_recv transfer
				if(type == DATA) {
					*recv_data = comp.buf;
				}
			}
		} else if (ret < 0 && ret != -FI_EAGAIN) {
			FT_PRINTERR("fi_cq_read", ret);
			return ret;
		}
	}
	return 0;
}

static int send_msg(int size)
{
	int ret;
	ctx_send = (struct fi_ctx_multi_recv *)
		malloc(sizeof(struct fi_ctx_multi_recv));
	ret = fi_send(ep, send_buf, (size_t) size, fi_mr_desc(mr),
			remote_fi_addr, &ctx_send->context);
	if (ret) {
		FT_PRINTERR("fi_send", ret);
		return ret;
	}

	ret = wait_for_send_completion(1);

	return ret;
}

static int sync_test(void)
{
	int ret;
	ret = opts.dst_addr ? send_msg(SYNC_DATA_SIZE) :
	   	wait_for_recv_completion(NULL, CONTROL, 1);
	if (ret)
		return ret;

	return opts.dst_addr ? wait_for_recv_completion(NULL, CONTROL, 1) :
		send_msg(SYNC_DATA_SIZE);
}

static int post_multi_recv_buffer()
{
	int ret = 0;
	int i;

	ctx_multi_recv = (struct fi_ctx_multi_recv *)
		malloc(sizeof(struct fi_ctx_multi_recv) * 2);

	// post buffers into halves so that we can
	// repost one half when the other half is full
	for(i = 0; i < 2; i++) {
		ctx_multi_recv[i].index = i;
		// post buffers for active messages
		ret = fi_recv(ep, multi_recv_buf + (multi_buf_size/2) * i,
			   	multi_buf_size/2, fi_mr_desc(mr_multi_recv),
				remote_fi_addr, &ctx_multi_recv[i].context);
		if (ret) {
			FT_PRINTERR("fi_recv", ret);
			return ret;
		}
	}

	return ret;
}

static int send_multi_recv_msg()
{
	int ret, i;
	ret = 0;
	// send multi_recv data based on the transfer size
	for(i = 0; i < opts.iterations; i++) {
		ctx_send = (struct fi_ctx_multi_recv *)
			malloc(sizeof(struct fi_ctx_multi_recv));
		ret = fi_send(ep, send_buf, (size_t) opts.transfer_size,
				fi_mr_desc(mr), remote_fi_addr,
				&ctx_send->context);
		if (ret) {
			FT_PRINTERR("fi_send", ret);
			return ret;
		}
		ret = wait_for_send_completion(1);
	}
	return ret;
}

static int run_test(void)
{
	int ret;

	ret = sync_test();
	if (ret) {
		fprintf(stderr, "sync_test failed!\n");
		goto out;
	}

	clock_gettime(CLOCK_MONOTONIC, &start);
	if(opts.dst_addr) {
		ret = send_multi_recv_msg();
		if(ret){
		  fprintf(stderr, "send_multi_recv_msg failed!\n");
			goto out;
		}
	} else {
		// wait for all the receive completion events for
		// multi_recv transfer
		ret = wait_for_recv_completion(NULL, CONTROL, opts.iterations);
		if(ret){
		  fprintf(stderr, "wait_for_recv_completion failed\n");
			goto out;
		}
	}

	clock_gettime(CLOCK_MONOTONIC, &end);

	if (opts.machr)
		show_perf_mr(opts.transfer_size, opts.iterations,
			&start, &end, 1, opts.argc, opts.argv);
	else
		show_perf(test_name, opts.transfer_size, opts.iterations,
			&start, &end, 1);

	ret = 0;

out:
	return ret;
}

static void free_res(void)
{
	FT_CLOSE_FID(mr_multi_recv);
	if (ctx_send) {
		free(ctx_send);
		ctx_send = NULL;
	}
	if (ctx_multi_recv) {
		free(ctx_multi_recv);
		ctx_multi_recv = NULL;
	}
	if (send_buf) {
		free(send_buf);
		send_buf = NULL;
	}
	if (multi_recv_buf) {
		free(multi_recv_buf);
		multi_recv_buf = NULL;
	}
}

static int alloc_ep_res(struct fi_info *fi)
{
	struct fi_cq_attr cq_attr;
	struct fi_av_attr av_attr;
	int ret;

	// size of the local address that needs to be exchanged later
	int data_size = sizeof(size_t);

	// maximum size of the buffer that needs to allocated
	max_send_buf_size = MAX(MAX(data_size, SYNC_DATA_SIZE), opts.transfer_size);

	if(max_send_buf_size > fi->ep_attr->max_msg_size) {
		fprintf(stderr, "transfer size is larger than the maximum size "
				"of the data transfer supported by the provider\n");
		return -1;
	}

	send_buf = malloc(max_send_buf_size);
	if (!send_buf) {
		fprintf(stderr, "Cannot allocate send_buf\n");
		return -1;
	}

	ret = fi_mr_reg(domain, send_buf, max_send_buf_size, 0, 0, 0, 0, &mr, NULL);
	if (ret) {
		FT_PRINTERR("fi_mr_reg", ret);
		return ret;
	}

	// set the multi buffer size to be allocated
	multi_buf_size = MAX(max_send_buf_size, DEFAULT_MULTI_BUF_SIZE) *
		MULTI_BUF_SIZE_FACTOR;
	multi_recv_buf = malloc(multi_buf_size);
	if (!multi_recv_buf) {
		fprintf(stderr, "Cannot allocate multi_recv_buf\n");
		return -1;
	}

	ret = fi_mr_reg(domain, multi_recv_buf, multi_buf_size, 0, 0, 1, 0,
			&mr_multi_recv, NULL);
	if (ret) {
		FT_PRINTERR("fi_mr_reg", ret);
		return ret;
	}

	memset(&cq_attr, 0, sizeof cq_attr);
	cq_attr.format = FI_CQ_FORMAT_DATA;
	cq_attr.wait_obj = FI_WAIT_NONE;
	cq_attr.size = max_credits << 1;
	ret = fi_cq_open(domain, &cq_attr, &txcq, NULL);
	if (ret) {
		FT_PRINTERR("fi_cq_open", ret);
		return ret;
	}

	ret = fi_cq_open(domain, &cq_attr, &rxcq, NULL);
	if (ret) {
		FT_PRINTERR("fi_cq_open", ret);
		return ret;
	}

	memset(&av_attr, 0, sizeof av_attr);
	av_attr.type = fi->domain_attr->av_type ?
			fi->domain_attr->av_type : FI_AV_MAP;
	av_attr.count = 1;
	av_attr.name = NULL;

	ret = fi_av_open(domain, &av_attr, &av, NULL);
	if (ret) {
		FT_PRINTERR("fi_av_open", ret);
		return ret;
	}

	ret = fi_endpoint(domain, fi, &ep, NULL);
	if (ret) {
		FT_PRINTERR("fi_endpoint", ret);
		return ret;
	}

	return 0;
}

static int set_min_multi_recv()
{
	int ret;

	ret = fi_setopt(&ep->fid, FI_OPT_ENDPOINT, FI_OPT_MIN_MULTI_RECV,
			(void *)&max_send_buf_size, sizeof(max_send_buf_size));
	if(ret)
		return ret;

	return 0;
}

static int init_fabric(void)
{
	uint64_t flags = 0;
	char *node, *service;
	int ret;

	ret = ft_read_addr_opts(&node, &service, hints, &flags, &opts);
	if (ret)
		return ret;

	ret = fi_getinfo(FT_FIVERSION, node, service, flags, hints, &fi);
	if (ret) {
		FT_PRINTERR("fi_getinfo", ret);
		return ret;
	}

	/* Get remote address */
	if (opts.dst_addr) {
		addrlen = fi->dest_addrlen;
		remote_addr = malloc(addrlen);
		memcpy(remote_addr, fi->dest_addr, addrlen);
	}

	// set FI_MULTI_RECV flag for all recv operations
	fi->rx_attr->op_flags = FI_MULTI_RECV;

	ret = ft_open_fabric_res();
	if (ret)
		return ret;

	ret = fi_domain(fabric, fi, &domain, NULL);
	if (ret) {
		FT_PRINTERR("fi_domain", ret);
		return ret;
	}

	ret = alloc_ep_res(fi);
	if (ret)
		return ret;

	ret = ft_init_ep(NULL);
	if (ret)
		return ret;

	// set the value of FI_OPT_MIN_MULTI_RECV which defines the minimum
	// receive buffer space available when the receive buffer is
	// automatically freed
	set_min_multi_recv();

	// post the initial receive buffers to get MULTI_RECV data
	ret = post_multi_recv_buffer();
	return ret;
}

static int init_av(void)
{
	int ret;
	void *recv_buf = NULL;

	if (opts.dst_addr) {
		// Get local address blob. Find the addrlen first. We set
		// addrlen as 0 and fi_getname will return the actual addrlen.
		addrlen = 0;
		ret = fi_getname(&ep->fid, local_addr, &addrlen);
		if (ret != -FI_ETOOSMALL) {
			FT_PRINTERR("fi_getname", ret);
			return ret;
		}

		local_addr = malloc(addrlen);
		ret = fi_getname(&ep->fid, local_addr, &addrlen);
		if (ret) {
			FT_PRINTERR("fi_getname", ret);
			return ret;
		}

		ret = fi_av_insert(av, remote_addr, 1, &remote_fi_addr, 0,
				&fi_ctx_av);
		if (ret != 1) {
			FT_PRINTERR("fi_av_insert", ret);
			return ret;
		}

		// Send local addr size
		memcpy(send_buf, &addrlen, sizeof(size_t));
		ret = send_msg(sizeof(size_t));
		if (ret)
			return ret;

		// Send local addr
		memcpy(send_buf, local_addr, addrlen);
		ret = send_msg(addrlen);
		if (ret)
			return ret;

	} else {

		// Get the size of remote address
		ret = wait_for_recv_completion(&recv_buf, DATA, 1);
		if (ret)
			return ret;
		memcpy(&addrlen, recv_buf, sizeof(size_t));

		// Get remote address
		remote_addr = malloc(addrlen);
		ret = wait_for_recv_completion(&recv_buf, DATA, 1);
		if (ret)
			return ret;
		memcpy(remote_addr, recv_buf, addrlen);

		ret = fi_av_insert(av, remote_addr, 1, &remote_fi_addr, 0,
				&fi_ctx_av);
		if (ret != 1) {
			FT_PRINTERR("fi_av_insert", ret);
			return ret;
		}
		ret = 0;
	}

	return ret;
}

static int run(void)
{
	int ret = 0;

	ret = init_fabric();
	if (ret)
		goto out;

	ret = init_av();
	if (ret)
		goto out;

	ret = run_test();
	/* Finalize before closing ep */
	ft_finalize(fi, ep, txcq, rxcq, remote_fi_addr);
out:
	return ret;
}

int main(int argc, char **argv)
{
	int op, ret;
	opts = INIT_OPTS;

	hints = fi_allocinfo();
	if (!hints)
		return EXIT_FAILURE;

	while ((op = getopt(argc, argv, "h" CS_OPTS INFO_OPTS)) != -1) {
		switch (op) {
		default:
			ft_parseinfo(op, optarg, hints);
			ft_parsecsopts(op, optarg, &opts);
			break;
		case '?':
		case 'h':
			ft_csusage(argv[0], "Ping pong client and server using message endpoints.");
			return EXIT_FAILURE;
		}
	}

	if (optind < argc)
		opts.dst_addr = argv[optind];

	hints->ep_attr->type = FI_EP_RDM;
	hints->caps = FI_MSG | FI_MULTI_RECV;
	hints->mode = FI_CONTEXT | FI_LOCAL_MR;

	ret = run();

	free_res();
	ft_free_res();
	return -ret;
}
